"""This module exists to hold global state for a Zino process"""

__all__ = ["polldevs", "ZinoState"]

import json
import logging
import os
from typing import Dict, Optional

from pydantic import BaseModel, Field

from zino.config.models import Configuration, IPAddress, PollDevice
from zino.events import Events
from zino.flaps import FlappingStates
from zino.planned_maintenance import PlannedMaintenances
from zino.statemodels import DeviceStates
from zino.utils import log_time_spent

_log = logging.getLogger(__name__)

# Dictionary of configured devices
polldevs: Dict[str, PollDevice] = {}

# Global (sic) state
state: "ZinoState" = None

config: Configuration = Configuration()

# Last time the pollfile was modified
pollfile_mtime: Optional[float] = None


class ZinoState(BaseModel):
    """Holds all state that Zino needs to persist between runtimes"""

    devices: DeviceStates = Field(default_factory=DeviceStates)
    events: Events = Field(default_factory=Events)
    addresses: dict[IPAddress, str] = {}
    planned_maintenances: PlannedMaintenances = Field(default_factory=PlannedMaintenances)
    flapping: FlappingStates = Field(default_factory=FlappingStates)

    @log_time_spent()
    def dump_state_to_file(self, filename: str):
        """Dumps the full state to a file in JSON format"""
        _log.debug("dumping state to %s", filename)
        temp_file = f"{filename}.tmp"
        with open(temp_file, "w") as statefile:
            statefile.write(self.model_dump_json(exclude_none=True, indent=2))
        os.replace(src=temp_file, dst=filename)

    @classmethod
    @log_time_spent()
    def load_state_from_file(cls, filename: str) -> Optional["ZinoState"]:
        """Loads and returns a previously persisted ZinoState from a JSON file dump.

        :returns: A ZinoState object if the state file was found, None if it wasn't.  If the state file is invalid or
                  otherwise unparseable, the underlying exceptions will be the callers responsibility to handle.
        """
        _log.info("Loading saved state from %s", filename)
        try:
            with open(filename, "r") as statefile:
                json_state = json.load(statefile)
            loaded_state = cls.model_validate(json_state)
        except FileNotFoundError:
            _log.error("No state file found (%s), starting from scratch ", filename)
            return
        else:
            return loaded_state



import pydantic
import pytest

from zino.config.models import PollDevice


class TestPollDevice:
    def test_init_should_fail_on_missing_address_or_name(self):
        with pytest.raises(pydantic.ValidationError):
            PollDevice()

    def test_init_should_succeed_with_address_and_name(self):
        assert PollDevice(name="example-gw", address="127.0.0.1")



import logging

from zino.state import ZinoState
from zino.stateconverter.linedata import LineData
from zino.stateconverter.utils import OldState

_log = logging.getLogger(__name__)


def set_flapping_state(old_state: OldState, new_state: ZinoState):
    for linedata in old_state["::firstFlap"]:
        _set_first_flap(linedata, new_state)
    for linedata in old_state["::flapHistVal"]:
        _set_flap_hist_val(linedata, new_state)
    for linedata in old_state["::flappedAboveThreshold"]:
        _set_flapped_above_threshold(linedata, new_state)
    for linedata in old_state["::flapping"]:
        _set_flapping(linedata, new_state)
    for linedata in old_state["::flaps"]:
        _set_flaps(linedata, new_state)
    for linedata in old_state["::lastFlap"]:
        _set_last_flap(linedata, new_state)
    for linedata in old_state["::lastFlap"]:
        _set_last_age(linedata, new_state)


def _set_first_flap(linedata: LineData, state: ZinoState):
    _log.info("firstFlap is not supported")


def _set_flap_hist_val(linedata: LineData, state: ZinoState):
    _log.info("flapHistVal is not supported")


def _set_flapped_above_threshold(linedata: LineData, state: ZinoState):
    _log.info("flappedAboveThreshold is not supported")


def _set_flapping(linedata: LineData, state: ZinoState):
    _log.info("flapping is not supported")


def _set_flaps(linedata: LineData, state: ZinoState):
    _log.info("flaps is not supported")


def _set_last_flap(linedata: LineData, state: ZinoState):
    _log.info("lastFlap is not supported")


def _set_last_age(linedata: LineData, state: ZinoState):
    """This is related to flapping"""
    _log.info("lastAge is is not supported")



import json
import os
from json import JSONDecodeError

import pytest

from zino.state import ZinoState


def test_dump_state_to_file_should_dump_valid_json_to_file(tmp_path):
    dumpfile = tmp_path / "dump.json"
    state = ZinoState()
    state.dump_state_to_file(dumpfile)

    assert os.path.exists(dumpfile)
    with open(dumpfile, "r") as data:
        assert json.load(data)


class TestLoadStateFromFile:
    def test_should_raise_on_invalid_json(self, invalid_state_file):
        with pytest.raises(JSONDecodeError):
            ZinoState.load_state_from_file(str(invalid_state_file))

    def test_should_load_saved_state(self, valid_state_file):
        state = ZinoState.load_state_from_file(str(valid_state_file))
        assert state.events.last_event_id == 42

    def test_should_return_none_when_state_file_is_missing(self, tmp_path):
        fake_file = tmp_path / "nonexistent.json"
        assert ZinoState.load_state_from_file(str(fake_file)) is None


#
# Fixtures
#


@pytest.fixture
def valid_state_file(tmp_path):
    state_filename = tmp_path / "dump.json"
    with open(state_filename, "w") as statefile:
        statefile.write(
            """
        {
          "devices": {
            "devices": {}
          },
          "events": {
            "events": {
              "1": {
                "id": 1,
                "router": "example-gw1",
                "type": "reachability",
                "state": "open",
                "opened": "2023-12-06T17:03:38.73336Z",
                "updated": "2023-12-06T17:03:38.733633Z",
                "priority": 100,
                "log": [
                  {
                    "timestamp": "2023-12-06T17:03:38.733633Z",
                    "message": "example-gw1 no-response"
                  }
                ],
                "history": [
                  {
                    "timestamp": "2023-12-06T17:03:38.733615Z",
                    "message": "Change state to Open"
                  }
                ],
                "reachability": "no-response"
              }
            },
            "last_event_id": 42
          }
        }
        """
        )
    return state_filename


@pytest.fixture
def invalid_state_file(tmp_path):
    state_filename = tmp_path / "invalid-dump.json"
    with open(state_filename, "w") as state_file:
        state_file.write("{....")
    return state_filename



import logging
from time import time
from unittest.mock import Mock, patch

import pytest
from apscheduler.jobstores.base import JobLookupError

from zino import scheduler
from zino.state import ZinoState


class TestLoadPolldevs:
    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_all_new_devices_on_first_run(self, polldevs_conf):
        new_devices, deleted_devices, changed_devices, _ = scheduler.load_polldevs(polldevs_conf)
        assert len(new_devices) > 0
        assert not deleted_devices
        assert not changed_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_defaults_on_first_run(self, polldevs_conf):
        _, _, _, defaults = scheduler.load_polldevs(polldevs_conf)
        assert len(defaults) > 0
        assert "interval" in defaults

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_deleted_devices_on_second_run(self, polldevs_conf, polldevs_conf_with_single_router):
        scheduler.load_polldevs(polldevs_conf)

        # This needs to be patched since the mtime of the two conf fixtures is the same
        with patch("zino.state.pollfile_mtime", last_run_time=time() - 60):
            new_devices, deleted_devices, changed_devices, _ = scheduler.load_polldevs(polldevs_conf_with_single_router)

        assert not new_devices
        assert len(deleted_devices) > 0
        assert not changed_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test__or_deleted_devices_on_invalid_configuration(self, invalid_polldevs_conf):
        new_devices, deleted_devices, changed_devices, _ = scheduler.load_polldevs(invalid_polldevs_conf)
        assert not new_devices
        assert not deleted_devices
        assert not changed_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_log_error_on_invalid_configuration(self, caplog, invalid_polldevs_conf):
        with caplog.at_level(logging.ERROR):
            scheduler.load_polldevs(invalid_polldevs_conf)
        assert "'lalala' is not a valid configuration line" in caplog.text

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_changed_defaults(self, polldevs_conf, tmp_path):
        polldevs_with_changed_defaults = tmp_path.joinpath("changed-defaults-polldevs.cf")
        with open(polldevs_with_changed_defaults, "w") as conf:
            conf.write(
                """# polldevs test config
                default interval: 10
                default community: barfoo
                default domain: uninett.no
                default statistics: yes
                default hcounters: yes

                name: example-gw
                address: 10.0.42.1

                name: example-gw2
                address: 10.0.43.1"""  # Lack of a new-line here is intentional to test the parser
            )

        _, _, _, defaults = scheduler.load_polldevs(polldevs_conf)

        # This needs to be patched since the mtime of the two conf fixtures is the same
        with patch("zino.state.pollfile_mtime", last_run_time=time() - 60):
            _, _, _, changed_defaults = scheduler.load_polldevs(polldevs_with_changed_defaults)
        assert defaults != changed_defaults

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_changed_devices_on_changed_defaults(self, polldevs_conf, tmp_path):
        polldevs_with_changed_defaults = tmp_path.joinpath("changed-defaults-polldevs.cf")
        with open(polldevs_with_changed_defaults, "w") as conf:
            conf.write(
                """# polldevs test config
                default interval: 10
                default community: barfoo
                default domain: uninett.no
                default statistics: yes
                default hcounters: yes

                name: example-gw
                address: 10.0.42.1

                name: example-gw2
                address: 10.0.43.1"""  # Lack of a new-line here is intentional to test the parser
            )

        scheduler.load_polldevs(polldevs_conf)

        # This needs to be patched since the mtime of the two conf fixtures is the same
        with patch("zino.state.pollfile_mtime", last_run_time=time() - 60):
            new_devices, deleted_devices, changed_devices, _ = scheduler.load_polldevs(polldevs_with_changed_defaults)

        assert not new_devices
        assert not deleted_devices
        assert changed_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_changed_devices_on_changed_interval(self, polldevs_conf, tmp_path):
        polldevs_with_changed_defaults = tmp_path.joinpath("changed-interval-polldevs.cf")
        with open(polldevs_with_changed_defaults, "w") as conf:
            conf.write(
                """# polldevs test config
                default interval: 5
                default community: foobar
                default domain: uninett.no
                default statistics: yes
                default hcounters: yes

                name: example-gw
                address: 10.0.42.1
                interval: 10

                name: example-gw2
                address: 10.0.43.1"""  # Lack of a new-line here is intentional to test the parser
            )

        scheduler.load_polldevs(polldevs_conf)

        # This needs to be patched since the mtime of the two conf fixtures is the same
        with patch("zino.state.pollfile_mtime", last_run_time=time() - 60):
            new_devices, deleted_devices, changed_devices, _ = scheduler.load_polldevs(polldevs_with_changed_defaults)

        assert not new_devices
        assert not deleted_devices
        assert changed_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_no_new_or_deleted_devices_on_unchanged_configuration(self, polldevs_conf):
        scheduler.load_polldevs(polldevs_conf)
        new_devices, deleted_devices, _, _ = scheduler.load_polldevs(polldevs_conf)
        assert not new_devices
        assert not deleted_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_return_no_new_or_deleted_devices_on_non_existent_pollfile(self, tmp_path):
        new_devices, deleted_devices, _, _ = scheduler.load_polldevs(tmp_path / "non-existent-polldev.cf")
        assert not new_devices
        assert not deleted_devices

    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_log_error_on_non_existent_pollfile(self, caplog, tmp_path):
        with caplog.at_level(logging.ERROR):
            scheduler.load_polldevs(tmp_path / "non-existent-polldev.cf")
        assert "No such file or directory" in caplog.text


class TestScheduleNewDevices:
    @patch("zino.state.polldevs", dict())
    @patch("zino.state.pollfile_mtime", None)
    @patch("zino.state.state", ZinoState())
    def test_should_schedule_jobs_for_new_devices(self, polldevs_conf, mocked_scheduler):
        new_devices, _, _, _ = scheduler.load_polldevs(polldevs_conf)
        assert len(new_devices) > 0

        scheduler.schedule_devices(new_devices)
        assert mocked_scheduler.add_job.called

    @patch("zino.state.polldevs", dict())
    def test_should_do_nothing_when_device_list_is_empty(self, mocked_scheduler):
        scheduler.schedule_devices([])
        assert not mocked_scheduler.add_job.called


def test_deschedule_deleted_devices_should_deschedule_jobs(mocked_scheduler):
    scheduler.deschedule_devices(["test-gw"])
    assert mocked_scheduler.remove_job.called


def test_deschedule_deleted_devices_should_not_fail_on_not_found_job(mocked_scheduler_raising_error, caplog):
    with caplog.at_level(logging.DEBUG):
        scheduler.deschedule_devices(["test-gw"])
    assert mocked_scheduler_raising_error.remove_job.called
    assert "Job for device test-gw could not be found" in caplog.text


def test_scheduler_should_be_initialized_without_error():
    sched = scheduler.get_scheduler()
    assert sched


@pytest.fixture
def mocked_scheduler():
    with patch("zino.scheduler.get_scheduler") as get_scheduler:
        mock_scheduler = Mock()
        get_scheduler.return_value = mock_scheduler

        yield mock_scheduler


@pytest.fixture
def mocked_scheduler_raising_error():
    with patch("zino.scheduler.get_scheduler") as get_scheduler:
        mock_scheduler = Mock()
        mock_scheduler.remove_job.side_effect = Mock(side_effect=JobLookupError(job_id="fake id"))
        get_scheduler.return_value = mock_scheduler

        yield mock_scheduler
